local addonName, addonTable = ...;

NIDisplay = {
    lastUpdateTime = 0;
}


local UPDATE_INTERVALL = 0.5; --[s]
local NIGHT_LENGTH = 0.6;
local FADETIME = 2;
local ALPHA_VARIANCE = 0.02;

local LIGHT_WIDTH = 2048; --[px]
local LIGHT_HEIGHT = 2048; --[px]

local illusionFrame;
local lightTexture;
local blindTexture;

local combatActive = false;

BINDING_HEADER_NIGHTILLUSION_BIND_HEADER   = NightIllusion_L["BIND_HEADER"]
BINDING_NAME_NIGHTILLUSION_BIND_DECREASE   = NightIllusion_L["BIND_DECREASE"]
BINDING_NAME_NIGHTILLUSION_BIND_INCREASE   = NightIllusion_L["BIND_INCREASE"]
BINDING_NAME_NIGHTILLUSION_BIND_RESET      = NightIllusion_L["BIND_AUTO"]
BINDING_NAME_NIGHTILLUSION_BIND_ILLUMINATE = NightIllusion_L["BIND_ILLUMINATE"]
BINDING_NAME_NIGHTILLUSION_BIND_DUNGEON    = NightIllusion_L["BIND_DUNGEON"]
BINDING_NAME_NIGHTILLUSION_BIND_ACTIVE     = NightIllusion_L["BIND_ACTIVE"]

----------------------------------------------------------------------------------------------------
-- EVENTS

function NIDisplay:OnLoad( frame )
    frame:RegisterEvent( "PLAYER_REGEN_DISABLED" ); -- combat enter
    frame:RegisterEvent( "PLAYER_REGEN_ENABLED"  ); -- combat leave
    frame:RegisterEvent( "PLAYER_LOGIN" ); -- login
    
    
    illusionFrame = getglobal("NightIllusion_IllusionFrame");
    lightTexture  = getglobal("NightIllusion_IllusionFrameLightTexture");
    blindTexture  = getglobal("NightIllusion_IllusionFrameBlindTexture");
end

function NIDisplay:OnEvent( frame, event, ... )
    local args = {...};
    -- NICore:PrintD("Event! " .. tostring(event));
    NICore:PrintD("EVENT Display! " .. tostring(event));
    
    if ( event == "PLAYER_REGEN_DISABLED" ) then
        if NISettings:IsCombatSafe() and NISettings:IsActive() then
            combatActive = NISettings:IsActive();
            NIDisplay:Disable()
        end
    elseif ( event == "PLAYER_REGEN_ENABLED" ) then
        if NISettings:IsCombatSafe() and combatActive then
            NIDisplay:Enable()
            combatActive = false;
        end
    elseif ( event == "PLAYER_LOGIN" ) then
        if not NISettings:IsLoaded() then
            NISettings:GetDefault();
        elseif NISettings:CheckUpdate() then
            NISettings:Update()
        end
        if NISettings:IsActive() then
            NIDisplay:AdjustAlpha();
            
            -- NIDisplay:SetAlpha(NISettings:GetDungeonAlpha())
            NIDisplay:RepositionLight(NISettings:GetLightPosition())
            NIDisplay:RescaleLight(NISettings:GetLightScaling())
        end
    end
end

function NIDisplay:OnUpdate( frame, elapsed )
    if not NISettings:IsActive() then
        return;
    end
    
    NIDisplay.lastUpdateTime = NIDisplay.lastUpdateTime + elapsed;    
    if (NIDisplay.lastUpdateTime > UPDATE_INTERVALL) then
        NIDisplay.lastUpdateTime = 0;
    else
        return;
    end
    
    if NISettings:IsDungeonMode() then
        NIDisplay:ShowIllusions();
        NIDisplay:SetAlpha( NISettings:GetDungeonAlpha() );
        -- NICore:PrintD("UPDATE! Dungeon: " ..tostring(NIDisplay:GetAlpha()) .. " = "  ..tostring(NISettings:GetDungeonAlpha()))
    elseif IsIndoors() then
        if NIDisplay:GetAlpha() > 0 + ALPHA_VARIANCE then 
            -- NICore:PrintD("UPDATE -> Inside, lighter")
            NIDisplay:Fade(NIDisplay:GetAlpha(), 0);            
        else
            NIDisplay:HideIllusions();
        end
        -- NICore:PrintD("UPDATE! Indoors: " .. tostring(NIDisplay:GetAlpha()));
    elseif NIDisplay:IsNight() then
        if NIDisplay:GetAlpha() < NIDisplay:CalculateAutoAlpha() - ALPHA_VARIANCE then
            -- NICore:PrintD("UPDATE -> darker")
            NIDisplay:Fade(NIDisplay:GetAlpha(), NIDisplay:CalculateAutoAlpha());
        elseif NIDisplay:GetAlpha() > NIDisplay:CalculateAutoAlpha() + ALPHA_VARIANCE then
            -- NICore:PrintD("UPDATE -> lighter")
            NIDisplay:Fade(NIDisplay:GetAlpha(), NIDisplay:CalculateAutoAlpha());
        end
        NIDisplay:ShowIllusions();
        -- NICore:PrintD("UPDATE! Night: ")
    else
        NIDisplay:HideIllusions();
        -- NICore:PrintD("UPDATE! Rest: ")
    end
end


----------------------------------------------------------------------------------------------------
-- DISPLAY functionality

function NIDisplay:Enable()
    NISettings:SetActive( true )
    NIDisplay:ShowIllusions()
    illusionFrame:Show();
end

function NIDisplay:Disable()
    NISettings:SetActive( false )
    NIDisplay:SetAlpha( 0 )
    NIDisplay:HideIllusions()
    illusionFrame:Hide();
end

function NIDisplay:IsEnabled()
    return NISettings:IsActive()
end

function NIDisplay:HideIllusions()
    NIDisplay:SetAlpha(0);
    
    lightTexture:Hide();
    blindTexture:Hide();
end

function NIDisplay:ShowIllusions()
    if NISettings:IsIlluminated() then
        lightTexture:Show();
        blindTexture:Hide();
    else
        blindTexture:Show();
        lightTexture:Hide();
    end
end

function NIDisplay:IsNight(now, startNight, endNight)
    if startNight== nil or endNight == nil or now == nil then
        startTh, startTm = NISettings:GetStartTime();
        endTh, endTm = NISettings:GetEndTime();
    
        local hh,mm = GetGameTime();
        startNight = NIDisplay:PrettyTime(startTh, startTm);
        endNight = NIDisplay:PrettyTime(endTh, endTm);
        now = NIDisplay:PrettyTime(hh,mm)
    end
    
    if NIDisplay:CompareTimeCircular(now, startNight, endNight) >= 0 then
        return true;
    else
        return false;
    end
end

-- function NIDisplay:IsAutoAlpha()
    -- return NIDisplay:IsNight() and not IsIndoors() and not NISettings:IsDungeonMode();
-- end

function NIDisplay:AdjustAlpha()
    local alpha = 0;
    
    if NISettings:IsAutomaticMode() then
        alpha = NIDisplay:CalculateAutoAlpha();
    else
        alpha = NISettings:GetDungeonAlpha();
    end
    NIDisplay:SetAlpha(alpha);
end

function NIDisplay:EnableDungeonMode()
    if not NISettings:IsDungeonMode() then
        NISettings:SetDungeonMode( true )
        NIDisplay:SetDungeonAlpha( NIDisplay:GetAlpha() )
    end
end

function NIDisplay:DisableDungeonMode()
    NISettings:SetDungeonMode( false )
end

function NIDisplay:SetDungeonAlpha( value )
    NIDisplay:EnableDungeonMode()
    NISettings:SetDungeonAlpha( value );
end

function NIDisplay:IncreaseDungeonAlpha( step )
    NIDisplay:EnableDungeonMode()
    
    local value = NISettings:GetDungeonAlpha() + step;
    NISettings:SetDungeonAlpha( value )
end

function NIDisplay:DecreaseDungeonAlpha( step )
    NIDisplay:EnableDungeonMode()
    
    local value = NISettings:GetDungeonAlpha() - step;
    NISettings:SetDungeonAlpha( value )
end

function NIDisplay:Fade( from, to ) -- 0: no fading, 1: get darker, -1: get lighter
    if from < to then
        -- NICore:PrintD("Fading: getting darker");
        local fadingAlpha = NIDisplay:GetAlpha() + ( ( to - from ) / FADETIME );
        
        if fadingAlpha > to then
            fadingAlpha = to
        end
        
        NIDisplay:SetAlpha(fadingAlpha);
        
        return 1
    elseif to < from then
        -- NICore:PrintD("Fading: getting lighter");
        local fadingAlpha = NIDisplay:GetAlpha() - ( (from - to ) / FADETIME );
        
        if fadingAlpha < to then
            fadingAlpha = to;
        end
        
        NIDisplay:SetAlpha(fadingAlpha)
    
        return -1
    else
        return 0
    end
end

function NIDisplay:SetAlpha(alpha)
    lightTexture:SetAlpha(alpha);
    blindTexture:SetAlpha(alpha);
end

function NIDisplay:GetAlpha()
    return blindTexture:GetAlpha();
end

function NIDisplay:CalculateAutoAlpha(nowT, startT, endT)
    local hh,mm = GetGameTime();
    local alpha = 0;
    if NIDisplay:IsNight() then -- IsAutoAlpha() then
        if nowT == nil or startT == nil or endT == nil then
            startT = NIDisplay:PrettyTime( NISettings:GetStartTime() ); 
            endT = NIDisplay:PrettyTime( NISettings:GetEndTime() );
            nowT = NIDisplay:PrettyTime( hh, mm );
        end
        
        local h, m = NIDisplay:CalculateTime("24:00", startT, "-");
        h, m = NIDisplay:CalculateTime(NIDisplay:PrettyTime(h,m), endT, "+");
        local nightLengthFull = NIDisplay:TransformHoursToMinutes(h,m);

        local nightLength = nightLengthFull * NIGHT_LENGTH;
        local dawnLength = floor( (nightLengthFull * (1 - NIGHT_LENGTH)) / 2 );
        local steps = 1 / dawnLength;
        
        local dawnH, dawnM = NIDisplay:TransformMinutesToHours(dawnLength);
        local dawnTime = NIDisplay:PrettyTime(dawnH, dawnM);
        
        local duskH, duskM = NIDisplay:CalculateTime(startT, dawnTime, "+");
        local dawnH, dawnM = NIDisplay:CalculateTime(endT, dawnTime, "-");
        local duskT = NIDisplay:PrettyTime(duskH, duskM);
        local dawnT = NIDisplay:PrettyTime(dawnH, dawnM);

        local startM = NIDisplay:TransformHoursToMinutes( NISettings:GetStartTime() );
        local endM = NIDisplay:TransformHoursToMinutes( NISettings:GetEndTime() );
        local nowM = NIDisplay:TransformHoursToMinutes( hh, mm );

        if ( NIDisplay:CompareTimeCircular(nowT, startT, duskT) >= 0 ) then
        -- Dusk
            local expiredH, expiredM = NIDisplay:CalculateTime(duskT, nowT, "-");
            local expired = NIDisplay:TransformHoursToMinutes(expiredH, expiredM);
            alpha = 1 - expired * steps;
        elseif ( NIDisplay:CompareTimeCircular(nowT, duskT, dawnT) >= 0 ) then
        -- Night
            alpha = 1.0;
        elseif ( NIDisplay:CompareTimeCircular(nowT, dawnT, endT) >= 0 ) then
        -- Dawn
            local expiredH, expiredM = NIDisplay:CalculateTime(endT, nowT, "-");
            local expired = NIDisplay:TransformHoursToMinutes(expiredH, expiredM);
            alpha = expired * steps;
        else
        -- Day
            alpha = 0.0;
        end
        
        
        if alpha > NISettings:GetMaxAlpha() then
            return NISettings:GetMaxAlpha();
        elseif alpha < NISettings:GetMinAlpha() then
            return NISettings:GetMinAlpha();
        else
            return alpha;
        end
    end
    
    return alpha;
end

function NIDisplay:ResetShadow()
    NISettings:SetDungeonMode( false );
    NIDisplay:AdjustAlpha( 0 );
end

function NIDisplay:GetTimes(startTime, endTime)
    if startTime == nil or endTime == nil then
        startTime = NIDisplay:PrettyTime( NISettings:GetStartTime() ); 
        endTime = NIDisplay:PrettyTime( NISettings:GetEndTime() );
    end
    
    local h, m = NIDisplay:CalculateTime("24:00", startTime, "-");
    h, m = NIDisplay:CalculateTime(NIDisplay:PrettyTime(h,m), endTime, "+");
    local nightLengthFull = NIDisplay:TransformHoursToMinutes(h,m);

    local nightLength = nightLengthFull * NIGHT_LENGTH;
    local dawnLength = floor( (nightLengthFull * (1 - NIGHT_LENGTH)) / 2 );
    local steps = 1 / dawnLength;
    
    local dawnH, dawnM = NIDisplay:TransformMinutesToHours(dawnLength);
    local dawnTime = NIDisplay:PrettyTime(dawnH, dawnM);
    
    local duskH, duskM = NIDisplay:CalculateTime(startTime, dawnTime, "+");
    local dawnH, dawnM = NIDisplay:CalculateTime(endTime, dawnTime, "-");
    local duskTime = NIDisplay:PrettyTime(duskH, duskM);
    local dawnTime = NIDisplay:PrettyTime(dawnH, dawnM);
    
    local duskT, nightT, dawnT, dayT;
    duskT = startTime;
    nightT = duskTime;
    dawnT = dawnTime;
    dayT = endTime;
    
    return duskT, nightT, dawnT, dayT;
end

local TIMEOFDAY_DUSK = "Dusk";
local TIMEOFDAY_NIGHT = "Night";
local TIMEOFDAY_DAWN = "Dawn";
local TIMEOFDAY_DAY = "Day";
local TIMEOFDAY_UNDEF = "UNDEF";
function NIDisplay:GetTimeOfDay(nowTime, startTime, endTime)
    local duskT, nightT, dawnT, dayT = NIDisplay:GetTimes(startTime, endTime);
    local tod = "";
    
    if NIDisplay:CompareTimeCircular(nowTime, duskT, nightT) >= 0 then
        tod = TIMEOFDAY_DUSK;
    elseif NIDisplay:CompareTimeCircular(nowTime, nightT, dawnT) >= 0 then
        tod = TIMEOFDAY_NIGHT;
    elseif NIDisplay:CompareTimeCircular(nowTime, dawnT, dayT) >= 0 then
        tod = TIMEOFDAY_DAWN;
    elseif NIDisplay:CompareTimeCircular(nowTime, dayT, duskT) >= 0 then
        tod = TIMEOFDAY_DAY;
    else
        tod = TIMEOFDAY_UNDEF;
    end
    
    return tod;
end

function NIDisplay:RepositionLight()
    local posX, posY = NISettings:GetLightPosition()
    local tex = lightTexture;
    local point, relativeTo, relativePoint, xOffset, yOffset = tex:GetPoint("CENTER")

    tex:SetPoint(point , relativeTo , relativePoint, posX, posY)
end

function NIDisplay:RescaleLight()
    local scaleX, scaleY = NISettings:GetLightScaling()
    local tex = lightTexture
    local w, h = NightIllusion_IllusionFrame:GetSize()
    
    local width = LIGHT_WIDTH * scaleX;
    local height = LIGHT_HEIGHT * scaleY;
    
    tex:SetSize(width, height);
end

function NIDisplay:ChangeColor(red, green, blue, alpha)
    if a == nil then
        a = 1.0
    end
    
    lightTexture:SetTexture(red, green, blue, alpha);
    blindTexture:SetTexture(red, green, blue, alpha);
end

----------------------------------------------------------------------------------------------------
-- Functions for time

-- returns hh, mm in 24h format; nil on error
function NIDisplay:Calc24H( timeString )
    if timeString == nil then
        return nil;
    end
    
    local hh, mm;
    
    local pTime = "^%d?%d:%d%d";
    local pMeridiem  = "[ap]%.?m%.?$";
    
    local mTime = strmatch(timeString, pTime);
    local mMeridiem = strmatch(timeString, pMeridiem);
    if mTime == nil then
        return nil, nil;
    end
    
    hh, mm = strsplit(":", mTime);
    
    hh = tonumber(hh);
    mm = tonumber(mm);
    if mm > 60 then
        hh = hh + ((mm - mod(mm, 60)) / 60);
        mm = mm - (mm - mod(mm, 60));
    end
    
    if hh < 0 or hh > 24 then
        return nil;
    end
    
    if mm < 0 or mm > 59 then
        return nil;
    end
    
    if strfind(tostring(mMeridiem), "^p") then
        hh = hh + 12;
    end
    
    return hh, mm
end

-- function NIDisplay:Calc12H( timeString ) --REM
-- end

function NIDisplay:PrettyTime( hh, mm, meri )
    local sTime = "";
    
    if hh == nil or mm == nil then
        return nil;
    end
    
    if meri ~= nil and strmatch(meri, "[ap]%.?m%.?$") then
        if hh - 12 > 0 then
            meri = " p.m.";
            hh = hh - 12;
        else
            meri = " a.m.";
        end
    elseif meri == nil then
        meri = "";
    end
    
    if mm > 59 then
        hh = hh + (mm - mod(mm, 60))/60;
        mm = mod(mm, 60);
    end
    if hh > 23 then
        hh = mod(hh, 24);
    end
    
    if mm < 10 then
        mm = "0" .. mm;
    end
    
    sTime = hh .. ":" .. mm .. "" .. meri;
    
    return sTime;
end

-- compares t1 to t2,  -1: t1 < t2; 0: t1==t2; 1: t1 > t2; nil error
function NIDisplay:CompareTime(t1, t2)
    t1 = NIDisplay:PrettyTime(NIDisplay:Calc24H( t1 ));
    t2 = NIDisplay:PrettyTime(NIDisplay:Calc24H( t2 ));
    
    if t1 == nil or t2 == nil then
        return nil
    end
    
    local h1, m1 = strsplit(":", t1);
    local h2, m2 = strsplit(":", t2);
    
    h1 = tonumber(h1);
    h2 = tonumber(h2);
    m1 = tonumber(m1);
    m2 = tonumber(m2);
        
    if h1 > h2 or ( h1 == h2 and m1 > m2 ) then
        return 1;
    elseif h1 < h2 or ( h1 == h2 and m1 < m2 ) then
        return -1;
    else
        return 0;
    end
end

-- compares t1 with [z1, z2]:  >=0: z1 <= t1 <= z2; -1: z1 > t1 < z2
function NIDisplay:CompareTimeCircular(t1, z1, z2)
    t1 = NIDisplay:PrettyTime(NIDisplay:Calc24H( t1 ));
    z1 = NIDisplay:PrettyTime(NIDisplay:Calc24H( z1 ));
    z2 = NIDisplay:PrettyTime(NIDisplay:Calc24H( z2 ));
    
    if t1 == nil or z1 == nil or z2 == nil then
        return nil
    end
        
    local th, tm = strsplit(":", t1);
    local h1, m1 = strsplit(":", z1);
    local h2, m2 = strsplit(":", z2);
    
    th = tonumber(th);
    tm = tonumber(tm);
    h1 = tonumber(h1);
    m1 = tonumber(m1);
    h2 = tonumber(h2);
    m2 = tonumber(m2);
    
    local hasMidnight = false;
    if NIDisplay:CompareTime(z1, z2) > 0 then
        hasMidnight = true;
    end
    
    if (th == h1 and tm == m1) and (th == h2 and tm == m2) then
        -- z1 == t1 == z2
        return 0;
    elseif  th == h1 and tm == m1 then
        -- t1 == z1
        return 1;
    elseif th == h2 and tm == m2  then
        -- t1 == z2
        return 2;
    elseif ( not hasMidnight and ( th > h1 or ( th == h1 and tm > m1 ) ) and ( th < h2 or ( th == h2 and tm < m2 ) ) ) 
            then
        -- z1 < t1 < z2
        return 3; 
    elseif ( hasMidnight and ( NIDisplay:CompareTimeCircular(t1, z1, "23:59") >= 0 )) then
        return 4;
    elseif ( hasMidnight and ( NIDisplay:CompareTimeCircular(t1, "00:00", z2) >= 0 )) then
        return 5;
    else
        return -1;
    end
end

function NIDisplay:TransformHoursToMinutes( hh, mm )
    return hh * 60 + mm;
end

function NIDisplay:TransformMinutesToHours( minutes )    
    local hh, mm;
    mm = minutes;
    hh = 0;
    if mm > 59 then
        hh = (mm - mod(mm, 60))/60;
        mm = mod(mm, 60);
    end
    
    if hh > 23 then
        hh = mod(hh, 24);
    end
    
    return hh, mm;
end

-- t1 op t2 = hh, mm
function NIDisplay:CalculateTime(t1, t2, op)
    local h1, m1 = NIDisplay:Calc24H( t1 );
    local h2, m2 = NIDisplay:Calc24H( t2 );
    
    local rh, rm;
    if op == "+" then
        rh = h1 + h2;
        rm = m1 + m2;
        
        rh = rh + (rm - mod(rm, 60))/60;
        rm = rm - (rm - mod(rm,60));        
    elseif op == "-" then
        rh = h1 - h2;
        rm = m1 - m2;
        
        local lostH = 0;
        while rm < 0 do
            rm = rm + 60;
            lostH = lostH + 1;
        end
        if rh < 0 then
            rh = 24 + rh;
        end
        rh = rh - lostH;
        
        rh = rh - (rm - mod(rm, 60))/60;
        rm = rm - (rm - mod(rm,60));    
    else
        return nil;
    end
    rh = mod(rh, 24);
    
    return rh, rm
end


function NIDisplay:PrintState()
    local state = "";
    if NISettings:IsActive() then
        state = "|cFF00FF00[ON]|r";
    else
        state = "|cFFFF0000[OFF]|r";
    end
    
    print("NightIllusion State: " .. state);
    print("    Inside: " .. tostring(IsIndoors()));
    print("    Vision: " .. tostring(lightTexture:IsVisible()) .. ", " .. tostring(lightTexture:GetAlpha()));
    print("    Blind: " .. tostring(blindTexture:IsVisible()) .. ", " .. tostring(blindTexture:GetAlpha()));
    print(" ");
    print("    MinAlpha: " .. tostring(NISettings:GetMinAlpha()));
    print("    MaxAlpha: " .. tostring(NISettings:GetMaxAlpha()));
    print("    Alpha: " .. tostring(NIDisplay:GetAlpha()));
    print("    StartNight: " .. NIDisplay:PrettyTime(NISettings:GetStartTime()));
    print("    EndNight: " .. NIDisplay:PrettyTime(NISettings:GetEndTime()));
    print(" ");
    print("    DungeonMode: " .. tostring(NISettings:IsDungeonMode()));
    print("    Illumination: " .. tostring(NISettings:IsIlluminated()) );
    print("    IsNight: " .. tostring(NIDisplay:IsNight()) );
    print("    ");
end